/*
 * JNpp - Java bindings for NPP, to be used with JCuda
 *
 * Copyright (c) 2010-2013 Marco Hutter - http://www.jcuda.org
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include "JNpp_common.hpp"

jfieldID cuComplex_x; // float
jfieldID cuComplex_y; // float

jfieldID cuDoubleComplex_x; // double
jfieldID cuDoubleComplex_y; // double

jfieldID NppiPoint_x; // int
jfieldID NppiPoint_y; // int

jfieldID NppiSize_width; // int
jfieldID NppiSize_height; // int

jfieldID NppiRect_x; // int
jfieldID NppiRect_y; // int
jfieldID NppiRect_width; // int
jfieldID NppiRect_height; // int


jfieldID Npp16sc_re; // short
jfieldID Npp16sc_im; // short

jfieldID Npp32sc_re; // int
jfieldID Npp32sc_im; // int

jfieldID Npp64sc_re; // long
jfieldID Npp64sc_im; // long

jfieldID Npp32fc_re; // float
jfieldID Npp32fc_im; // float

jfieldID Npp64fc_re; // double
jfieldID Npp64fc_im; // double


// Class and method ID for TypedPointer and its constructor
jclass TypedPointer_Class;
jmethodID TypedPointer_Constructor;

/**
 * Creates a global reference to the class with the given name and
 * stores it in the given jclass argument, and stores the no-args
 * constructor ID for this class in the given jmethodID.
 * Returns whether this initialization succeeded.
 */
bool init(JNIEnv *env, const char *className, jclass &globalCls, jmethodID &constructor)
{
    jclass cls = NULL;
    if (!init(env, cls, className)) return false;
    if (!init(env, cls, constructor, "<init>", "()V")) return false;

    globalCls = (jclass)env->NewGlobalRef(cls);
    if (globalCls == NULL)
    {
        Logger::log(LOG_ERROR, "Failed to create reference to class %s\n", className);
        return false;
    }
    return true;
}

/**
 * Called when the library is loaded. Will initialize all
 * required field and method IDs
 */
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved)
{
    JNIEnv *env = NULL;
    if (jvm->GetEnv((void **)&env, JNI_VERSION_1_4))
    {
        return JNI_ERR;
    }

    Logger::log(LOG_TRACE, "Initializing JNpp\n");

    jclass cls = NULL;

    // Initialize the JNIUtils and PointerUtils
    if (initJNIUtils(env) == JNI_ERR) return JNI_ERR;
    if (initPointerUtils(env) == JNI_ERR) return JNI_ERR;

    // Obtain the constructor for the TypedPointer class
    if (!init(env, "jcuda/jnpp/TypedPointer", TypedPointer_Class, TypedPointer_Constructor)) return JNI_ERR;

    // Obtain the fieldIDs for cuComplex#x and cuComplex#y
    if (!init(env, cls, "jcuda/cuComplex")) return JNI_ERR;
    if (!init(env, cls, cuComplex_x, "x", "F")) return JNI_ERR;
    if (!init(env, cls, cuComplex_y, "y", "F")) return JNI_ERR;

    // Obtain the fieldIDs for cuDoubleComplex#x and cuDoubleComplex#y
    if (!init(env, cls, "jcuda/cuDoubleComplex")) return JNI_ERR;
    if (!init(env, cls, cuDoubleComplex_x, "x", "D")) return JNI_ERR;
    if (!init(env, cls, cuDoubleComplex_y, "y", "D")) return JNI_ERR;

    // Obtain the fieldIDs for NppiPoint
    if (!init(env, cls, "jcuda/jnpp/NppiPoint")) return JNI_ERR;
    if (!init(env, cls, NppiPoint_x, "x", "I")) return JNI_ERR;
    if (!init(env, cls, NppiPoint_y, "y", "I")) return JNI_ERR;

    // Obtain the fieldIDs for NppiSize
    if (!init(env, cls, "jcuda/jnpp/NppiSize")) return JNI_ERR;
    if (!init(env, cls, NppiSize_width,  "width",  "I")) return JNI_ERR;
    if (!init(env, cls, NppiSize_height, "height", "I")) return JNI_ERR;

    // Obtain the fieldIDs for NppiRect
    if (!init(env, cls, "jcuda/jnpp/NppiRect")) return JNI_ERR;
    if (!init(env, cls, NppiRect_x,      "x",      "I")) return JNI_ERR;
    if (!init(env, cls, NppiRect_y,      "y",      "I")) return JNI_ERR;
    if (!init(env, cls, NppiRect_width,  "width",  "I")) return JNI_ERR;
    if (!init(env, cls, NppiRect_height, "height", "I")) return JNI_ERR;

    // Obtain the fieldIDs for Npp16sc
    if (!init(env, cls, "jcuda/jnpp/types/Npp16sc")) return JNI_ERR;
    if (!init(env, cls, Npp16sc_re, "re", "S")) return JNI_ERR;
    if (!init(env, cls, Npp16sc_im, "im", "S")) return JNI_ERR;

    // Obtain the fieldIDs for Npp32sc
    if (!init(env, cls, "jcuda/jnpp/types/Npp32sc")) return JNI_ERR;
    if (!init(env, cls, Npp32sc_re, "re", "I")) return JNI_ERR;
    if (!init(env, cls, Npp32sc_im, "im", "I")) return JNI_ERR;

    // Obtain the fieldIDs for Npp64sc
    if (!init(env, cls, "jcuda/jnpp/types/Npp64sc")) return JNI_ERR;
    if (!init(env, cls, Npp64sc_re, "re", "J")) return JNI_ERR;
    if (!init(env, cls, Npp64sc_im, "im", "J")) return JNI_ERR;

    // Obtain the fieldIDs for Npp32fc
    if (!init(env, cls, "jcuda/jnpp/types/Npp32fc")) return JNI_ERR;
    if (!init(env, cls, Npp32fc_re, "re", "F")) return JNI_ERR;
    if (!init(env, cls, Npp32fc_im, "im", "F")) return JNI_ERR;

    // Obtain the fieldIDs for Npp64fc
    if (!init(env, cls, "jcuda/jnpp/types/Npp64fc")) return JNI_ERR;
    if (!init(env, cls, Npp64fc_re, "re", "D")) return JNI_ERR;
    if (!init(env, cls, Npp64fc_im, "im", "D")) return JNI_ERR;

    return JNI_VERSION_1_4;

}


/**
 * Create a new TypedPointer object with the given native pointer value
 */
jobject createPointer(JNIEnv *env, jlong pointerValue)
{
    jobject result = env->NewObject(TypedPointer_Class, TypedPointer_Constructor);
    if (env->ExceptionCheck())
    {
        return NULL;
    }
    setPointer(env, result, pointerValue);
    return result;
}


/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, NppiPoint &output)
{
    output.x = (int)env->GetIntField(input, NppiPoint_x);
    output.y = (int)env->GetIntField(input, NppiPoint_y);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, NppiSize &output)
{
    output.width  = (int)env->GetIntField(input, NppiSize_width);
    output.height = (int)env->GetIntField(input, NppiSize_height);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, NppiRect &output)
{
    output.x      = (int)env->GetIntField(input, NppiRect_x);
    output.y      = (int)env->GetIntField(input, NppiRect_y);
    output.width  = (int)env->GetIntField(input, NppiRect_width);
    output.height = (int)env->GetIntField(input, NppiRect_height);
    return true;
}


//=== Complex number struct init =============================================

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, Npp16sc &output)
{
    output.re = (Npp16s)env->GetShortField(input, Npp16sc_re);
    output.im = (Npp16s)env->GetShortField(input, Npp16sc_im);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, Npp32sc &output)
{
    output.re = (Npp32s)env->GetIntField(input, Npp32sc_re);
    output.im = (Npp32s)env->GetIntField(input, Npp32sc_im);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, Npp64sc &output)
{
    output.re = (Npp64s)env->GetLongField(input, Npp64sc_re);
    output.im = (Npp64s)env->GetLongField(input, Npp64sc_im);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, Npp32fc &output)
{
    output.re = (Npp32f)env->GetFloatField(input, Npp32fc_re);
    output.im = (Npp32f)env->GetFloatField(input, Npp32fc_im);
    return true;
}

/**
 * Initializes the given native output object with the values from the
 * given java input object.
 */
bool initNative(JNIEnv *env, jobject input, Npp64fc &output)
{
    output.re = (Npp64f)env->GetDoubleField(input, Npp64fc_re);
    output.im = (Npp64f)env->GetDoubleField(input, Npp64fc_im);
    return true;
}



//=== Complex number struct release ==========================================

/**
 * Write the values from the given native output object back to
 * the java input object
 */
bool releaseNative(JNIEnv *env, Npp16sc &output, jobject input)
{
    env->SetShortField(input, Npp16sc_re, (jshort)output.re);
    env->SetShortField(input, Npp16sc_im, (jshort)output.im);
    return true;
}

/**
 * Write the values from the given native output object back to
 * the java input object
 */
bool releaseNative(JNIEnv *env, Npp32sc &output, jobject input)
{
    env->SetIntField(input, Npp32sc_re, (jint)output.re);
    env->SetIntField(input, Npp32sc_im, (jint)output.im);
    return true;
}

/**
 * Write the values from the given native output object back to
 * the java input object
 */
bool releaseNative(JNIEnv *env, Npp64sc &output, jobject input)
{
    env->SetLongField(input, Npp64sc_re, (jlong)output.re);
    env->SetLongField(input, Npp64sc_im, (jlong)output.im);
    return true;
}

/**
 * Write the values from the given native output object back to
 * the java input object
 */
bool releaseNative(JNIEnv *env, Npp32fc &output, jobject input)
{
    env->SetFloatField(input, Npp32fc_re, (jfloat)output.re);
    env->SetFloatField(input, Npp32fc_im, (jfloat)output.im);
    return true;
}

/**
 * Write the values from the given native output object back to
 * the java input object
 */
bool releaseNative(JNIEnv *env, Npp64fc &output, jobject input)
{
    env->SetDoubleField(input, Npp64fc_re, (jdouble)output.re);
    env->SetDoubleField(input, Npp64fc_im, (jdouble)output.im);
    return true;
}





//=== 2D prmitive arrays =====================================================


// TODO: Many of these methods could be replaced by more generic
// ones that should be defined in CommonJNI

/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, double output[4][2])
{
    int outerSize = 4;
    int innerSize = 2;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jdoubleArray inputElement = (jdoubleArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jdouble *arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            output[i][j] = (double)arrayContents[j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, JNI_ABORT);
    }
    return true;
}

/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, double output[4][2], jobjectArray input)
{
    int outerSize = 4;
    int innerSize = 2;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jdoubleArray inputElement = (jdoubleArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jdouble *arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            arrayContents[j] = (jdouble)output[i][j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, 0);
    }
    return true;
}





/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, double output[2][3])
{
    int outerSize = 2;
    int innerSize = 3;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jdoubleArray inputElement = (jdoubleArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jdouble *arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            output[i][j] = (double)arrayContents[j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, JNI_ABORT);
    }
    return true;
}

/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, double output[2][3], jobjectArray input)
{
    int outerSize = 2;
    int innerSize = 3;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jdoubleArray inputElement = (jdoubleArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jdouble *arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            arrayContents[j] = (jdouble)output[i][j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, 0);
    }
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp32f output[3][4])
{
    int outerSize = 3;
    int innerSize = 4;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jfloatArray inputElement = (jfloatArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jfloat *arrayContents = (jfloat*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            output[i][j] = (Npp32f)arrayContents[j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, JNI_ABORT);
    }
    return true;
}

/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32f output[3][4], jobjectArray input)
{
    int outerSize = 3;
    int innerSize = 4;
    jsize size = env->GetArrayLength(input);
    if (size < outerSize)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<outerSize; i++)
    {
        jfloatArray inputElement = (jfloatArray)env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException", "Array contains 'null' elements");
            return false;
        }
        jsize elementSize = env->GetArrayLength(inputElement);
        if (elementSize < innerSize)
        {
            ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException", "Array parameter has insufficient size");
            return false;
        }
        jfloat *arrayContents = (jfloat*)env->GetPrimitiveArrayCritical(inputElement, NULL);
        for (int j=0; j<innerSize; j++)
        {
            arrayContents[j] = (jfloat)output[i][j];
        }
        env->ReleasePrimitiveArrayCritical(inputElement, arrayContents, 0);
    }
    return true;
}






//=== Primitive arrays =======================================================

/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jbyteArray input, Npp8s output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jbyte* arrayContents = (jbyte*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp8s)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp8s output[], jbyteArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jbyte* arrayContents = (jbyte*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jbyte)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jbyteArray input, Npp8u output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jbyte* arrayContents = (jbyte*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp8u)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp8u output[], jbyteArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jbyte* arrayContents = (jbyte*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jbyte)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jshortArray input, Npp16s output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jshort* arrayContents = (jshort*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp16s)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp16s output[], jshortArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jshort* arrayContents = (jshort*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jshort)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jshortArray input, Npp16u output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jshort* arrayContents = (jshort*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp16u)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp16u output[], jshortArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jshort* arrayContents = (jshort*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jshort)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jintArray input, Npp32s output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jint* arrayContents = (jint*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp32s)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32s output[], jintArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jint* arrayContents = (jint*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jint)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jintArray input, Npp32u output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jint* arrayContents = (jint*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp32u)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32u output[], jintArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jint* arrayContents = (jint*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jint)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jlongArray input, Npp64s output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jlong* arrayContents = (jlong*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp64s)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp64s output[], jlongArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jlong* arrayContents = (jlong*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jlong)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jlongArray input, Npp64u output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jlong* arrayContents = (jlong*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp64u)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp64u output[], jlongArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jlong* arrayContents = (jlong*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jlong)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jfloatArray input, Npp32f output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jfloat* arrayContents = (jfloat*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp32f)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32f output[], jfloatArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jfloat* arrayContents = (jfloat*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jfloat)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jdoubleArray input, Npp64f output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jdouble* arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        output[i] = (Npp64f)arrayContents[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, JNI_ABORT);
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp64f output[], jdoubleArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    jdouble* arrayContents = (jdouble*)env->GetPrimitiveArrayCritical(input, NULL);
    if (arrayContents == NULL)
    {
        return false;
    }
    for (int i=0; i<size; i++)
    {
        arrayContents[i] = (jdouble)output[i];
    }
    env->ReleasePrimitiveArrayCritical(input, arrayContents, 0);
    return true;
}





//=== Complex number arrays ==================================================

/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp16sc output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!initNative(env, inputElement, output[i]))
        {
            return false;
        }
    }
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp16sc output[], jobjectArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!releaseNative(env, output[i], inputElement))
        {
            return false;
        }
    }
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp32sc output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!initNative(env, inputElement, output[i]))
        {
            return false;
        }
    }
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32sc output[], jobjectArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!releaseNative(env, output[i], inputElement))
        {
            return false;
        }
    }
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp32fc output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!initNative(env, inputElement, output[i]))
        {
            return false;
        }
    }
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp32fc output[], jobjectArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!releaseNative(env, output[i], inputElement))
        {
            return false;
        }
    }
    return true;
}



/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp64fc output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!initNative(env, inputElement, output[i]))
        {
            return false;
        }
    }
    return true;
}


/**
 * Release the given native output array, writing its contents back to
 * the given java input array
 */
bool releaseNative(JNIEnv *env, Npp64fc output[], jobjectArray input, int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        if (!releaseNative(env, output[i], inputElement))
        {
            return false;
        }
    }
    return true;
}





//=== Primitive pointer arrays ===============================================

/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp8u* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp8u*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp8u* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp8u*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp16s* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp16s*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp16s* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp16s*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp16u* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp16u*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp16u* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp16u*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp32s* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp32s*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp32s* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp32s*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp32f* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp32f*)getPointer(env, inputElement);
    }
    return true;
}




/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp32f* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp32f*)getPointer(env, inputElement);
    }
    return true;
}

/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, Npp64f* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (Npp64f*)getPointer(env, inputElement);
    }
    return true;
}


/**
 * Initialize the given native output array by filling it with the data from
 * the given java input array
 */
bool initNative(JNIEnv *env, jobjectArray input, const Npp64f* output[], int size)
{
    jsize arraySize = env->GetArrayLength(input);
    if (arraySize < size)
    {
        ThrowByName(env, "java/lang/ArrayIndexOutOfBoundsException",
            "Array parameter has insufficient size");
        return false;
    }
    for (int i=0; i<size; i++)
    {
        jobject inputElement = env->GetObjectArrayElement(input, i);
        if (inputElement == NULL)
        {
            ThrowByName(env, "java/lang/NullPointerException",
                "Input array contains 'null' elements");
            return false;
        }
        output[i] = (const Npp64f*)getPointer(env, inputElement);
    }
    return true;
}


